use core::convert::TryInto;
use core::ffi::c_void;
use core::slice::from_raw_parts;

use std::ffi::CString;
use std::os::raw::{c_uint, c_char};
use std::path::Path;

use libretro_sys::{Message, PixelFormat, SystemAvInfo, GameGeometry};
use num_enum::{IntoPrimitive, TryFromPrimitive};

use crate::libretro_convert::*;
use crate::libretro_loading::*;

static mut CB_SINGLETON: StaticCallbacks = StaticCallbacks {
    handler: None,
};

#[derive(Default)]
struct StaticCallbacks {
    handler: Option<&'static mut dyn Handler>,
}

unsafe impl Sync for StaticCallbacks {}

impl StaticCallbacks {
    // helpers for environ cb
    fn clone_into_void<T: Clone>(data: *mut c_void, source: &T) -> Option<bool> {
        unsafe { (data as *mut T).as_mut() }?.clone_from(source);
        Some(true)
    }
    fn string_into_void(data: *mut c_void, source: impl AsRef<str>) -> Option<bool> {
        *unsafe { (data as *mut *const c_char).as_mut()? } = CString::new(source.as_ref()).ok()?.into_raw();
        Some(true)
    }
    fn path_into_void(data: *mut c_void, source: impl AsRef<Path>) -> Option<bool> {
        Self::string_into_void(data, source.as_ref().to_string_lossy())
    }
    fn from_void<T>(data: *mut c_void) -> Option<&'static mut T> {
        unsafe { (data as *mut T).as_mut() }
    }
    fn enum_from_void<T>(data: *mut c_void) -> Option<T>
        where T: TryFromPrimitive, <T as TryFromPrimitive>::Primitive: 'static
    {
        let number = Self::from_void(data).cloned()?;
        T::try_from_primitive(number).ok()
    }
    fn environment_cb_inner(cmd: u32, data: *mut c_void) -> Option<bool> {
        let mut handler = unsafe { CB_SINGLETON.handler.as_mut() }?;
        match cmd.try_into().ok()? {
            EnvCmd::SetRotation => handler.set_rotation(Self::enum_from_void(data)?),
            EnvCmd::GetOverscan => Self::clone_into_void(data, &handler.get_overscan()?)?,
            EnvCmd::GetCanDupe => Self::clone_into_void(data, &handler.get_can_dupe()?)?,
            EnvCmd::SetMessage => handler.set_message(Self::from_void::<Message>(data)?.clone()),
            EnvCmd::Shutdown => handler.shutdown(),
            EnvCmd::SetPerformanceLevel => handler.set_performance_level(*Self::from_void(data)?),
            EnvCmd::GetSystemDirectory => Self::path_into_void(data, handler.get_system_directory()?)?,
            EnvCmd::SetPixelFormat => handler.set_pixel_format(PixelFormat::from_uint(*Self::from_void(data)?)?),
            // TODO EnvCmd::SetInputDescriptors => {},
            // TODO EnvCmd::SetKeyboardCallback => {},
            // TODO EnvCmd::SetDiskControlInterface => {},
            // TODO EnvCmd::SetHwRender => {},
            // TODO EnvCmd::GetVariable => {}, -- also change to mut parameter?
            // TODO EnvCmd::SetVariables => {},
            EnvCmd::GetVariableUpdate => Self::clone_into_void(data, &handler.get_variable_update()?)?,
            EnvCmd::SetSupportNoGame => handler.set_support_no_game(*Self::from_void(data)?),
            EnvCmd::GetLibretroPath => Self::path_into_void(data, handler.get_libretro_path()?)?,
            // TODO EnvCmd::SetFrameTimeCallback => {},
            // TODO EnvCmd::SetAudioCallback => {},
            // TODO EnvCmd::GetRumbleInterface => {},
            EnvCmd::GetInputDeviceCapabilities => Self::clone_into_void(data, &handler.get_input_device_capabilities()?)?,
            // TODO EnvCmd::GetSensorInterface => {},
            // TODO EnvCmd::GetCameraInterface => {},
            // TODO EnvCmd::GetLogInterface => {},
            // TODO EnvCmd::GetPerfInterface => {},
            // TODO EnvCmd::GetLocationInterface => {},
            EnvCmd::GetCoreAssetsDirectory => Self::path_into_void(data, handler.get_core_assets_directory()?)?,
            EnvCmd::GetSaveDirectory => Self::path_into_void(data, handler.get_save_directory()?)?,
            EnvCmd::SetSystemAvInfo => handler.set_system_av_info(Self::from_void::<SystemAvInfo>(data)?.clone()),
            // TODO EnvCmd::SetProcAddressCallback => {},
            // TODO EnvCmd::SetSubsystemInfo => {},
            // TODO EnvCmd::SetControllerInfo => {},
            // TODO EnvCmd::SetMemoryMaps => {},
            EnvCmd::SetGeometry => handler.set_geometry(Self::from_void::<GameGeometry>(data)?.clone()),
            EnvCmd::GetUsername => Self::string_into_void(data, handler.get_username()?)?,
            EnvCmd::GetLanguage => Self::clone_into_void(data, &handler.get_language()?)?,
            // EnvCmd::SetSerializationQuirks => handler.set_serialization_quirks(Self::from_void(data)?),
            _ => false,
        }.into()
    }
    extern "C" fn environment_cb(cmd: u32, data: *mut c_void) -> bool {
        Self::environment_cb_inner(cmd, data).unwrap_or(false)
    }

    extern "C" fn video_refresh_cb(
        data: *const c_void,
        width: c_uint,
        height: c_uint,
        pitch: usize,
    ) {
        if !data.is_null() {
            if let Some(cb) = unsafe { CB_SINGLETON.handler.as_mut() } {
                let data = data as *const u8;
                let len = pitch * (height as usize);
                let slice = unsafe { from_raw_parts(data, len) };
                cb.video_refresh(slice, width, height, pitch as u32);
            }
        }
    }
    extern "C" fn audio_sample_cb(left: i16, right: i16) {
        if let Some(cb) = unsafe { CB_SINGLETON.handler.as_mut() } {
            cb.audio_sample(left, right);
        }
    }
    extern "C" fn audio_sample_batch_cb(data: *const i16, frames: usize) -> usize {
        unsafe {
            match CB_SINGLETON.handler.as_mut() {
                Some(cb) => match data.is_null() {
                    true => 0,
                    false => cb.audio_sample_batch(from_raw_parts(data, frames)),
                },
                None => 0,
            }
        }
    }
    extern "C" fn input_poll_cb() {
        unsafe { CB_SINGLETON.handler.as_mut().map(|cb| cb.input_poll()) };
    }
    extern "C" fn input_state_cb(port: c_uint, device: c_uint, index: c_uint, id: c_uint) -> i16 {
        match unsafe { CB_SINGLETON.handler.as_mut() } {
            Some(cb) => match ((device, id).try_into(), index.try_into()) {
                (Ok(input), Ok(index)) => cb.input_state(port, input, index),
                _ => 0,
            },
            None => 0,
        }
    }
    pub(crate) fn set_environment(cb: &'static mut dyn Handler) {
        unsafe {
            CB_SINGLETON.handler.replace(cb);
        }
    }
    pub(crate) fn reset() {
        unsafe {
            CB_SINGLETON.handler.take();
        }
    }
}

pub struct LibretroWrapper<'a> {
    api: LibretroApi<'a>,
}

impl<'a> From<LibretroApi<'a>> for LibretroWrapper<'a> {
    fn from(api: LibretroApi<'a>) -> Self {
        api.set_environment(StaticCallbacks::environment_cb);
        api.set_video_refresh(StaticCallbacks::video_refresh_cb);
        api.set_audio_sample(StaticCallbacks::audio_sample_cb);
        api.set_audio_sample_batch(StaticCallbacks::audio_sample_batch_cb);
        api.set_input_poll(StaticCallbacks::input_poll_cb);
        api.set_input_state(StaticCallbacks::input_state_cb);
        LibretroWrapper { api }
    }
}

impl<'a> AsRef<LibretroApi<'a>> for LibretroWrapper<'a> {
    fn as_ref(&self) -> &LibretroApi<'a> {
        &self.api
    }
}

impl<'a> Drop for LibretroWrapper<'a> {
    fn drop(&mut self) {
        StaticCallbacks::reset();
    }
}

// a note on lifetimes: we explicitly lie about them here because as long as they live as long as
//  the library wrapper itself we're good (we wipe our 'static references on drop() too)
impl<'a> LibretroWrapper<'a> {
    pub fn register_handler(&self, handler: &'a mut dyn Handler) {
        let ptr: *mut dyn Handler = handler;
        StaticCallbacks::set_environment(unsafe { ptr.as_mut() }.unwrap());
    }
}
